/*
 *    Copyright  1996, 1997, 1998, 1999, 2000 by Remco Treffkorn
 *    Distributed under the terms of the Q Public License. Please see
 *    the included file LICENSE.
 */
#include "config.h"

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <math.h>

#include <X11/Xos.h>
#include <Xm/Text.h>
#include <Xm/TextF.h>
#include <Xm/LabelG.h>
#include <X11/xpm.h>

#include "xmap.h"
#include "mode.h"
#include "debug.h"

extern Widget toplevel;

#define MAXPOINTS 500

enum {
    O_PIXMAP, O_POSLABEL, O_TEXTLABEL, O_DRAWN, O_HIGHLIGHT, O_COMPLEX, O_WAYPOINT, O_ROUTE
};

#define O_NONE     0
#define O_SAVE     1
#define O_NEW      2
#define O_CHANGED  4
#define O_MARKER   8


struct d_obj {
    short type;			// Describes the type of object.

    short status;		// Current status, i.e. CHANGED, NEW...

    struct d_obj *next;		// Pointer to next object, if any.

    struct d_obj *prev;		// Pointer to previous object, if any.

    double lat;			// Object hot spot latitude

    double lon;			// Object hot spot longitude

    Position x;			// X related to raster map origin

    Position y;			// Y related to raster map origin

    int pmap;			// The objects pixmap index, if any.

    int n;			// Actual elements in drawlist

    int max;			// Max elements in drawlist

    void *drawlist;		// List of draw commands, if any.

    void *private;		// draw functions private data, if any.

};

struct d_obj *root = NULL;

extern Widget draww;

static int mx, my;
static struct d_obj *complexo;
static struct d_obj *routeo;

WP *get_WP()
{
    WP *WPe;
    if (!(WPe = (WP *)malloc(sizeof(WP)))) {
        error_wait("Can't make new waypoint!\n",errno); // exits
    }
    return WPe;
}

Coord *get_Coord()
{
    Coord *Ce;
    if (!(Ce = (Coord *)malloc(sizeof(Coord)))) {
        error_wait("Can't make new object!\n",errno); // exits
    }
    return Ce;
}

static void draw_pixmap(struct d_obj *o)
{
    Position x, y;

    latlontoxy(o->lat, o->lon, &x, &y);
    draw_object_pixmap(o->pmap, x, y);
}

static void draw_label(struct d_obj *o)
{
}

static void draw_text(struct d_obj *o)
{
}

static void draw_complex(struct d_obj *o)
{
    int i;
    XPoint *xpoints;

    i = o->n;
    xpoints = (XPoint *)malloc((i+1)*sizeof(XPoint));
    if (!xpoints)
	return;

    if (o->type == O_ROUTE) {
	while (i--) {
	    latlontoxy(((WP *) (o->drawlist))[i].lat, ((WP *) (o->drawlist))[i].lon,
		       &xpoints[i].x, &xpoints[i].y);
	}
	draw_route(draww, xpoints, o->n, True);
    } else {
	while (i--) {
	    latlontoxy(((Coord *) (o->drawlist))[i].lat, ((Coord *) (o->drawlist))[i].lon,
		       &xpoints[i].x, &xpoints[i].y);
	}
	highlight(draww, xpoints, o->n);
    }
    free((char *)xpoints);
}

void undraw_route(struct d_obj *o)
{
    int i;
    XPoint *xpoints;

    xpoints = (XPoint *)malloc((i+1)*sizeof(XPoint));
    if (!xpoints)
	return;

    i = o->n;

    while (i--) {
	latlontoxy(((WP *) (o->drawlist))[i].lat, ((WP *) (o->drawlist))[i].lon,
		   &xpoints[i].x, &xpoints[i].y);
    }
    draw_route(draww, xpoints, o->n, False);
}

void draw_object(struct d_obj *o)
{
    switch (o->type) {
    case O_PIXMAP:
    case O_WAYPOINT:
	draw_pixmap(o);
	break;

    case O_POSLABEL:
    case O_TEXTLABEL:
	draw_label(o);
	break;

    case O_DRAWN:
    case O_HIGHLIGHT:
    case O_COMPLEX:
    case O_ROUTE:
	draw_complex(o);
	break;
    }
}


static struct d_obj *create_object()
{
    struct d_obj *o, *p;

    if ( (o = (struct d_obj *) calloc(1, sizeof(struct d_obj))) ) {
	p = root;

	if (p) {
	    while (p->next)
		p = p->next;
	    p->next = o;
	    o->prev = p;
	} else
	    root = o;
    }
    return o;
}

static void destroy_object(struct d_obj *o)
{
    if (o->type == O_ROUTE) {
	int i;
	WP *w;

	i = o->n;
	while (i--) {
	    w = &((WP *) (o->drawlist))[i];
	    if (w->name)
		free(w->name);
	    if (w->description)
		free(w->description);
	    if (w->action)
		free(w->action);
	}
    }
    if (o->drawlist)
	free(o->drawlist);

    if (o->private)
	free(o->private);

    if (o->prev) {
	if (o->next) {
	    (o->next)->prev = o->prev;
	    o->prev = o->next;
	} else
	    (o->prev)->next = NULL;
    } else {
	if ((root = o->next))
	    root->prev = NULL;
    }

    free(o);
}

static void clear_objects(int all)
{
    struct d_obj *o, *n;

    o = root;

    ch_mode(MARK, STOP);
    while (o) {
	n = o->next;
	if (o->type != O_WAYPOINT && o->type != O_ROUTE)
	    if (all || (o->status&O_MARKER)) destroy_object(o);
	o = n;
    }
    refresh(0);
}

void clear_all_objects()
{
    clear_objects(True);
}

void clear_marker_objects()
{
    clear_objects(False);
}

void create_pixmap_object(double lat, double lon, int pmap)
{
    struct d_obj *o;

    if ( (o=create_object()) ) {
	o->status = O_NEW;
	o->type = O_PIXMAP;
	o->pmap = pmap;
	o->lat = lat;
	o->lon = lon;
	latlontoxy(lat, lon, &o->x, &o->y);

	draw_object(o);
    }
}

#if 0
void create_waypoint_object(double lat, double lon, int pmap, char *name, char *desc)
{
    struct d_obj *o;

    if (o = create_object()) {
	o->status = O_NEW;
	o->type = O_WAYPOINT;
	o->pmap = pmap;
	o->lat = lat;
	o->lon = lon;
	o->private = calloc(1, sizeof(struct wp_private));
	((struct wp_private *) o->private)->name = strdup(name);
	((struct wp_private *) o->private)->desc = strdup(desc);
	latlontoxy(lat, lon, &o->x, &o->y);
	draw_object(o);
    }
}

#endif

void redraw_objects(void)
{
    struct d_obj *o;

    o = root;

    if (o && mapdata.calibrated) {
	do {
	    draw_object(o);
	} while ( (o=o->next) );
    }
}

void do_draw()
{
    if (!mapdata.calibrated) {
	error("Can't draw on uncalibrated maps.\nCalibrate the map first.", 0);
	return;
    }
    if (!mode_is(MARK)) {
	if ((complexo = create_object())) {
	    if ((complexo->drawlist = calloc(MAXPOINTS, sizeof(Coord)))) {
		complexo->status = O_NEW;
		complexo->type = O_COMPLEX;
		complexo->n = 0;
		ch_mode(MARK, START);
		mx = my = -1;
		set_cursor(0);
	    }
	}
    } else {
	undraw_rubber(draww);
	ch_mode(MARK, STOP);
	set_cursor(0);
	draw_object(complexo);
    }
}

void *do_route()
{
    if (!mapdata.calibrated) {
	error("Can't route on uncalibrated maps.\nCalibrate the map first.", 0);
	return NULL;
    }
    if (!mode_is(CROUTE)) {
	if ((routeo = create_object())) {
	    if ((routeo->drawlist = calloc(MAXPOINTS, sizeof(WP)))) {
		routeo->status = O_NEW;
		routeo->max = MAXPOINTS;
		routeo->type = O_ROUTE;
		routeo->n = 0;
		ch_mode(CROUTE, START);
		mx = my = -1;
		set_cursor(0);
		return (void *) routeo;
	    }
	}
    } else {
	undraw_rubber(draww);
	ch_mode(CROUTE, STOP);
	set_cursor(0);
	draw_object(routeo);
    }
    return (void *) NULL;
}

void *start_route()
{
    if ((routeo = create_object())) {
	if ((routeo->drawlist = malloc(MAXPOINTS * sizeof(WP)))) {
	    routeo->status = O_SAVE;
	    routeo->type = O_ROUTE;
	    routeo->n = 0;
	    routeo->max = MAXPOINTS;
	    return (void *) routeo;
	}
    }
    return (void *) NULL;
}

int d_route_nelements(struct d_obj *croute)
{
    return croute->n;
}

WP *d_route_p(int n)
{
    if (routeo->n > n)
	return &((WP *) (routeo->drawlist))[n];
    else
	return NULL;
}

WP *wp_add()
{
    if (routeo->n >= routeo->max) {	// need to get more memory...
	routeo->drawlist = realloc(routeo->drawlist, (routeo->max + MAXPOINTS) * sizeof(WP));
	routeo->max += MAXPOINTS;
//	routeo->n--;	// just rewind the clock
    }
    memset( &((WP *) (routeo->drawlist))[routeo->n], 0, sizeof(WP));
    return &((WP *) (routeo->drawlist))[routeo->n++];
}

void route_delete(struct d_obj *croute)
{
    destroy_object(croute);
    refresh(0);
}

void route(int x, int y)
{
    xytolatlon(x, y, &((WP *) (routeo->drawlist))[routeo->n].lat,
	       &((WP *) (routeo->drawlist))[routeo->n].lon);
    routeo->n++;
    if (mx >= 0) {
	undraw_rubber(draww);
	draw_object(routeo);
    }
    mx = x;
    my = y;
}

void route_translate_rubber(int x, int y)
{
    int n;
    Position sx, sy;

    if ((n = routeo->n) > 0) {
	n--;
	latlontoxy(
		      ((WP *) (routeo->drawlist))[n].lat,
		      ((WP *) (routeo->drawlist))[n].lon,
		      &sx, &sy);
	mx = sx;
	my = sy;
	mark_move(x, y);
    }
}

int is_modified(struct d_obj *croute)
{
    return ((croute->status&O_NEW) || (croute->status&O_CHANGED));
}

void is_now_saved(struct d_obj *croute)
{
    croute->status |= O_SAVE;
}

void is_now_changed(struct d_obj *croute)
{
    croute->status |= O_CHANGED;
}

void mark_move(int x, int y)
{
    if (mx > 0) {
	undraw_rubber(draww);
	draw_rubber(draww, mx, my, x, y);
    }
}

void mark(int x, int y)
{
    xytolatlon(x, y, &((Coord *) (complexo->drawlist))[complexo->n].lat,
	       &((Coord *) (complexo->drawlist))[complexo->n].lon);
    complexo->n++;
    if (mx >= 0) {
	undraw_rubber(draww);
	draw_object(complexo);
    }
    mx = x;
    my = y;
}

static void save_label(struct d_obj *o, FILE * fp)
{
}

static void save_pixmap(struct d_obj *o, FILE * fp)
{
    char bufp[21];

    fprintf(fp, "pixmap % 2d\n", o->pmap);
    toRawLatDMM(o->lat, bufp, 21);
    fprintf(fp, "\t%s", bufp);
    toRawLonDMM(o->lon, bufp, 21);
    fprintf(fp, "%s\n", bufp);
}

static void save_complex(struct d_obj *o, FILE * fp)
{
    int i;
    char bufp[21];

    i = o->n;
    fprintf(fp, "complex %d\n", i);

    while (i--) {
	toRawLatDMM(((Coord *) (o->drawlist))[i].lat, bufp, 21);
	fprintf(fp, "\t%s", bufp);
	toRawLatDMM(((Coord *) (o->drawlist))[i].lon, bufp, 21);
	fprintf(fp, "%s\n", bufp);
    }
}

static void save_object(struct d_obj *o, FILE * fp)
{
    if (o->status&O_MARKER) return;

    switch (o->type) {
    case O_PIXMAP:
	save_pixmap(o, fp);
	break;

    case O_POSLABEL:
    case O_TEXTLABEL:
	save_label(o, fp);
	break;

    case O_DRAWN:
    case O_HIGHLIGHT:
    case O_COMPLEX:
	save_complex(o, fp);
	break;
    }
}

/* This is a poor way of doing this, but should work for now */
void save_waypoints()
{
    struct d_obj *o;
    char bufp[256];
    char wn[8];
    unsigned int ordinal;
    int i;

    ordinal = GetPrivateProfileInt("Waypoints:default", "ordinal", 0, inifile);

    o = root;
    i = 0;

    if (o) {
	do {
	    if (o->type == O_WAYPOINT) {
		sprintf(wn, "w%04d", i);
		sprintf(bufp, "%.6f %.6f %s,%s,", o->lat, o->lon,
			((struct wp_private *) o->private)->name,
			((struct wp_private *) o->private)->desc);
		WritePrivateProfileString("Waypoints:default", wn, bufp, 
					  inifile);
		i++;
	    }
	} while (o = o->next);
    }
    WritePrivateProfileInt("Waypoints:default", "ordinal", i, inifile);
}

void save_objects()
{
    FILE *fp;
    struct d_obj *o;
    char fname[256];

    sprintf(fname, "%s/hugo.obj", get_mapdir());

    if ((fp = fopen(fname, "w")) == NULL) {
	error("Can't open the object file!\nThe objects can not be saved!", errno);
	return;
    }
    fprintf(fp, "hugo.obj 1\n");
    o = root;

    if (o) {
	do {
	    save_object(o, fp);
	} while ( (o=o->next) );
    }
    fclose(fp);
    save_waypoints();
}

static void getcoords(FILE * fp, double *lat, double *lon)
{
    double m, d;
    char p, q;

    fscanf(fp, " %lf%c%lf%c\n", lat, &p, lon, &q);

    *lat /= 100.0;
    m = 100.0 * modf(*lat / 100.0, &d);
    *lat = d + m / 60.0;

    *lon /= 100.0;
    m = 100.0 * modf(*lon / 100.0, &d);
    *lon = d + m / 60.0;

    if (p == 'S')
	*lat = -(*lat);
    if (q == 'W')
	*lon = -(*lon);
}

static void read_complex(FILE * fp, int n, int flag)
{
    struct d_obj *o;
    double lat, lon;
    int l;

    if ((o = create_object())) {
	if ((o->drawlist = calloc(n, sizeof(Coord)))) {
	    o->status = flag;
	    o->type = O_COMPLEX;
	    o->n = n;
	}
    }
    for (l = 0; l < n; l++) {
	getcoords(fp, &lat, &lon);
	((Coord *) (o->drawlist))[l].lat = lat;
	((Coord *) (o->drawlist))[l].lon = lon;
    }
}


static void read_pixmap(FILE * fp, int n, int flag)
{
    struct d_obj *o;
    double lat, lon;

    if ((o = create_object())) {
	if ((o->drawlist = calloc(1, sizeof(Coord)))) {
	    o->status = flag;
	    o->type = O_PIXMAP;
	    o->pmap = n;
	}
    }
    getcoords(fp, &lat, &lon);
    o->lat = lat;
    o->lon = lon;
}

static void sread(char *fname, int flag)
{
    FILE *fp;
    int n;
    char bufp[256];

    if ((fp = fopen(fname, "r")) == NULL) {
	error("Can't read the object file", errno);
	return;
    }
    fgets(bufp, 255, fp);
    if (strcmp(bufp, "hugo.obj 1\n")) {
	error("Can't read the object file.\nIncorrect version.\n", 0);
	return;
    }
    while (fgets(bufp, 255, fp)) {
	if (strncmp(bufp, "complex", 7) == 0) {
	    sscanf(bufp, "complex %d", &n);
	    if (n>0) read_complex(fp, n, flag);
	}
	if (strncmp(bufp, "pixmap", 6) == 0) {
	    sscanf(bufp, "pixmap  %d", &n);
	    if (n<0 || n>=NUMOBJ) {
		sprintf(bufp, "Invalid pixmap index %d!\nNothing will be drawn.\n", n);
		error(bufp, 0);
	    } else
		read_pixmap(fp, n, flag);
	}
    }
    fclose(fp);
}

void read_objects()
{
    char fname[256];

    sprintf(fname, "%s/hugo.obj", get_mapdir());
    sread(fname, O_NONE);
}


void reload_objects()
{
    clear_all_objects();
    read_objects();
    redraw_objects();
}

void import_markers()
{
    select_marker_file(draww);
}

void marker_read(char *file)
{
    sread(file, O_MARKER);
}

#ifdef DND
void TransferProc(widget, client_data, seltype, type, value, length, format)
Widget widget;
XtPointer client_data;
Atom *seltype;
Atom *type;
XtPointer value;
unsigned long *length;
int format;
{
    Display *dpy;
    Atom PIXMAP_INDEX;
    Widget w;
    XmString string;
    char *label[256];

    return;
    /* intern the Atoms for data targets */
    dpy = XtDisplay(toplevel);
    PIXMAP_INDEX = XmInternAtom(dpy, "PIXMAP_INDEX", False);

    w = (Widget) client_data;

    if (*type == PIXMAP_INDEX)
	fprintf(stderr, "Hallo: %s\n", value);
}
#endif


#ifdef DND
void DropDraw(Widget w, XtPointer client_data, XtPointer call_data)
{
    Display *dpy;
    Atom PIXMAP_INDEX;
    XmDropProcCallback DropData;
    XmDropTransferEntryRec transferEntries[1];
    XmDropTransferEntry transferList;
    Arg args[10];
    int n, i;
    Widget dc;
    Cardinal numExportTargets;
    Atom *exportTargets;
    Boolean pixmap_index = False;
    void TransferProc();

    /* intern the Atoms for data targets */
    dpy = XtDisplay(toplevel);
    PIXMAP_INDEX = XmInternAtom(dpy, "PIXMAP_INDEX", False);

    DropData = (XmDropProcCallback) call_data;
    dc = DropData->dragContext;

    /* retrieve the data targets and search for FILE_CONTENTS */
    n = 0;
    XtSetArg(args[n], XmNexportTargets, &exportTargets);
    n++;
    XtSetArg(args[n], XmNnumExportTargets, &numExportTargets);
    n++;
    XtGetValues(dc, args, n);

    for (i = 0; i < numExportTargets; i++) {
	if (exportTargets[i] == PIXMAP_INDEX) {
	    pixmap_index = True;
	    break;
	}
    }

    if (pixmap_index) {
	/* make sure we have a drop that is a copy operation.
	 * if not, set the status to failure.
	 */
	n = 0;
	if ((DropData->dropAction != XmDROP) ||
	    (DropData->operation != XmDROP_COPY)) {
	    XtSetArg(args[n], XmNtransferStatus, XmTRANSFER_FAILURE);
	    n++;
	    XtSetArg(args[n], XmNnumDropTransfers, 0);
	    n++;
	} else {
	    /* set up transfer requests for drop site */
	    transferEntries[0].target = PIXMAP_INDEX;
	    transferList = transferEntries;
	    XtSetArg(args[n], XmNdropTransfers, transferList);
	    n++;
	    XtSetArg(args[n], XmNnumDropTransfers, 1);
	    n++;
	    XtSetArg(args[n], XmNtransferProc, TransferProc);
	    n++;
	}
	XmDropTransferStart(dc, args, n);
    } else
	beep();

    fprintf(stderr, "DROP!\n");
}
#endif
